from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

import arrow


@dataclass
class BaseData:
    """Base Data Class.

    The subclasses created using this class will be available
    to Data Scientists.

    Default Properties
    ==================

    * tax_id (str):                 The TaxId prospected.
                                    must be a valid CNPJ, formatted or not.
    * integration_status (str):     The current Integration Status. Current status are:
                                    NOT_FOUND, WAITING, SUCCESS, EXPIRED, ERROR,
                                    CONN_ERROR and TIMEOUT
                                    Please check the `find_current_status` method
                                    for more info.
    * status_reason (str):          The human friendly reason for current status
    * prospect_date (datetime):     The prospect datetime.
    * payload (dict):               Original payload received from prospection.
    * timeit                        Time in seconds for retrieving data.
    * datasource_base_url           Data Source Base Url
    """

    tax_id: Optional[str] = None
    integration_status: Optional[str] = None
    status_reason: str = ""
    prospect_date: Optional[datetime] = None
    payload: Optional[Dict[str, Any]] = None
    timeit: Optional[float] = None
    datasource_base_url: Optional[str] = None

    def find_current_status(self, max_days_old: int):
        """Find current Status for Data.

        Current available status:

        * NOT_FOUND   When backend responds 404 for TaxId
        * WAITING     When backend responds with status "PENDING" or "RUNNING"
        * SUCCESS     When backend responds with status "SUCCESS"
                        and prospect date is lower or equal than max_days_old
        * EXPIRED     When backend responds with status "SUCCESS"
                        and prospect date is greater than max_days_old
        * ERROR       When backend responds a integration error
                        when prospecting the DataSource
        * CONN_ERROR  When backend responds 40x,50x
        * TIMEOUT     When timeout occurs

        :param max_days_old: Max days old for current data, for check expiration.
        """
        if self.integration_status == "SUCCESS":
            days_old = (arrow.utcnow() - arrow.get(self.prospect_date)).days
            self.integration_status = (
                "EXPIRED" if days_old > max_days_old else "SUCCESS"
            )
        if self.integration_status in ["PENDING", "RUNNING"]:
            self.integration_status = "WAITING"



from dataclasses import dataclass
from typing import Any, Dict, Optional, Type

import arrow
import requests
from loguru import logger
from marshmallow import Schema
from requests import HTTPError, Timeout
from validate_docbr import CNPJ

from snatch.base.data import BaseData
from snatch.helpers.only_numbers import only_numbers
from snatch.helpers.time_it import time_it


@dataclass
class BaseSnatch:
    """Base Snatch Class.

    For each Datasource, you`ll need to subclass this class
    overriding the following properties:

    * serializer_class (Schema):    The Serializer Class used to deserialize
                                    backend response dict to DataSource Data Object.
                                    Must be a Marshmallow schema.
    * default_timeout (int):        Default timeout in seconds, for wait backend
                                    response for pending prospections.
                                    Please make sure you give enough
                                    time to backend complete prospect the Datasource.
    * default_max_days (int):       Default max days for expiration check. All
                                    prospections with prospect_date older than
                                    max_days will be marked as EXPIRED.
    * authorization_token (str):    The Datasource Authorization Token.
    * tax_id (str):                 The TaxId prospected. Must be a valid CNPJ,
                                    formatted or not.
    * base_url (str):               DataSource base url.
    """

    serializer_class: Optional[Type[Schema]] = None
    default_timeout: int = 60
    default_max_days: int = 30
    authorization_token: Optional[str] = None
    tax_id: Optional[str] = None
    base_url: Optional[str] = None

    @property
    def headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
            "Authorization": f"Token {self.authorization_token}",
        }

    @property
    def tax_id(self) -> str:  # noqa
        """Return current TaxId."""
        return self._tax_id

    @tax_id.setter
    def tax_id(self, data: str):
        """Save current TaxId for DataSource Integration.

        Will check a valid CNPJ first and convert to numbers only.

        :param data: TaxId (must be a valid CNPJ)
        :return: None
        """
        cnpj = CNPJ()
        if not cnpj.validate(data):
            raise ValueError("Invalid TaxId")
        self._tax_id = only_numbers(data)  # noqa

    def _waiter(self, timeout: Optional[int] = None):
        if not timeout:
            timeout = self.default_timeout
        max_waiting_time = arrow.utcnow().shift(seconds=timeout)
        now = arrow.utcnow()
        next_check = now
        while now <= max_waiting_time:
            if now >= next_check:
                next_check = arrow.utcnow().shift(seconds=1)
                data = self._last_prospect(timeout=timeout)
                if data and data.integration_status in [
                    "SUCCESS",
                    "ERROR",
                    "CONN_ERROR",
                    "NOT_FOUND",
                ]:
                    logger.info(f"Data Received - Status is: {data.integration_status}")
                    return data
            now = arrow.utcnow()
        return self._format_timeout_response("/last_prospect")

    def _post_new_prospect(self, timeout: int) -> BaseData:
        url = "/api/prospects/"
        payload = {"tax_id": self.tax_id}
        data = self._request_and_create_data(url=url, timeout=timeout, payload=payload)
        return data

    def _last_prospect(self, timeout: int) -> BaseData:
        url = f"/api/companies/{self.tax_id}/last_prospect/"
        return self._request_and_create_data(url=url, timeout=timeout)

    def _last_valid(self, timeout: int) -> BaseData:
        url = f"/api/companies/{self.tax_id}/last_valid/"
        return self._request_and_create_data(url=url, timeout=timeout)

    def _format_response(self, data: Dict[Any, Any]) -> BaseData:
        return self.serializer_class().load(data)

    def _format_error_response(self, error):
        status_code = (
            error.response.status_code if hasattr(error, "response") else "N/D"
        )
        integration_status = "NOT_FOUND" if status_code == 404 else "CONN_ERROR"
        return self.serializer_class().load(
            {
                "status_reason": f"Status Code: {status_code} "
                f"when accessing {error.request.url}",
                "integration_status": integration_status,
                "payload": error.response.json() if hasattr(error, "response") else {},
            }
        )

    def _format_timeout_response(self, url):
        return self.serializer_class().load(
            {
                "status_reason": f"Timeout when accessing: {url}",
                "integration_status": "TIMEOUT",
            }
        )

    def _request_and_create_data(
        self, url: str, timeout: int, payload: Dict[Any, Any] = None
    ):
        try:
            method = "POST" if payload else "GET"
            full_url = f"{self.base_url}{url}"
            logger.info(f"Start {method} request to {full_url} ...")
            response = requests.request(
                method=method,
                url=full_url,
                json=payload,
                headers=self.headers,
                timeout=timeout,
            )
            logger.debug(f"Response status code is: {response.status_code}")
            response.raise_for_status()
            return self._format_response(response.json())
        except HTTPError as error:
            return self._format_error_response(error)
        except Timeout as error:
            return self._format_timeout_response(error.request.url)

    @time_it
    def last_valid(self, timeout: Optional[int] = None):
        """Return last valid DataSource integration for TaxId.

        It will check for the last valid Datasource Integration
        for selected TaxId. The last valid integration is the
        last one with "SUCCESS" status, regardless prospect time.

        :param timeout: Time in seconds before timeout connection. Default: 60 seconds.
        :return: DataSource Data Object
        """
        if not timeout:
            timeout = self.default_timeout
        return self._last_valid(timeout=timeout)

    @time_it
    def get_data(
        self, timeout: Optional[int] = None, max_days_old: Optional[int] = None
    ):
        """Get or Request DataSource Integration Data.

        It will:

        1. Check if exists a valid integration for
        selected TaxId in backend. A valid integration
        is a successfully integration (status: SUCCESS) with
        prospect date less or equal the `max_days_old` informed.

        2. If integration does not exist,
        has EXPIRED or has an status "ERROR",
        automatically start a new integration.

        3. If integration was started or already PENDING/RUNNING,
        wait `timeout` seconds for the backend response.

        :param max_days_old: Max days old for last integration,
                                if exists. Default: 30 days old.
        :param timeout: Time in seconds before timeout connection. Default: 60 seconds.
        :return: DataSource Data Object
        """
        if not timeout:
            timeout = self.default_timeout

        if not max_days_old:
            max_days_old = self.default_max_days

        data = self._last_prospect(timeout=timeout)
        data.find_current_status(max_days_old=max_days_old)

        # Check Status
        if data.integration_status in ["SUCCESS", "TIMEOUT", "CONN_ERROR"]:
            return data

        if data.integration_status == "WAITING":
            wait_data = self._waiter(timeout=timeout)
            return wait_data

        # Status: NOT_FOUND, ERROR or EXPIRED
        self._post_new_prospect(timeout)
        wait_data = self._waiter(timeout=timeout)
        return wait_data



from marshmallow import EXCLUDE, Schema, fields


class BaseSchema(Schema):
    """Base DataSource Schema.

    Add a subclass for each Datasource and his specific fields.

    """

    tax_id = fields.Str(allow_none=True)
    prospect_date = fields.DateTime(allow_none=True, data_key="prospect_end_date")
    integration_status = fields.Str(data_key="status")
    status_reason = fields.Str(default="")

    class Meta:
        unknown = EXCLUDE



"""Get Environment Secrets for Snatch."""
import os
from typing import Any, Dict

from dotenv import load_dotenv
from loguru import logger
from scalpl import Cut

from snatch.helpers.secrets_manager import SecretManager

load_dotenv()


def get_environment_from_secrets_manager() -> Dict[Any, Any]:
    """Load settings from Secrets Manager..

    :return Dict[Any, Any]
    """
    environment = os.getenv("ENV", "local")

    secrets = SecretManager(
        current_environment=environment,
        project_name="snatch",
        current_env_data={},
    )

    if not secrets.can_read_secrets:
        return {}

    logger.info(
        f"Reading settings from Secrets Manager for "
        f"service: snatch and environment: {environment} ..."
    )

    secrets_manager_data = secrets.get_project_secrets()

    return secrets_manager_data


config = {
    "snatch": {
        "boa_vista_secret_token": os.getenv("SNATCH_BOA_VISTA_SECRET_TOKEN", "foo"),
        "datasource_boa_vista_url": os.getenv(
            "SNATCH_DATASOURCE_BOA_VISTA_URL", "http://datasource_boa_vista"
        ),
    },
    "project": {"aws_region": os.getenv("PROJECT_AWS_REGION", "us-east-1")},
}
config.update(get_environment_from_secrets_manager())

settings = Cut(config)



import json
import os
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import boto3
from asbool import asbool
from botocore.config import Config
from loguru import logger
from stela.utils import merge_dicts


@dataclass
class SecretManager:
    current_environment: str
    project_name: str
    current_env_data: Dict[Any, Any]
    _client: Optional[boto3.client] = None

    @property
    def can_read_secrets(self) -> bool:
        return (
            asbool(os.getenv("PROJECT_READ_SSM_DATA"))
            if os.getenv("PROJECT_READ_SSM_DATA")
            else True
        )

    @property
    def client(self) -> boto3.client:
        if self._client:
            return self._client

        if not os.getenv("AWS_ACCESS_KEY_ID_SHARED", None):
            # just initialize, will check for `.aws/credentials` file
            self._client = boto3.client("secretsmanager")
            return self._client

        aws_config = Config(region_name=self.current_env_data["project"]["aws_region"])
        client_config = {
            "config": aws_config,
            "aws_access_key_id": os.environ["AWS_ACCESS_KEY_ID_SHARED"],
            "aws_secret_access_key": os.environ["AWS_SECRET_ACCESS_KEY_SHARED"],
        }
        aws_session_token = os.getenv("AWS_SESSION_TOKEN_SHARED")

        if aws_session_token:
            client_config["aws_session_token"] = aws_session_token

        self._client = boto3.client("secretsmanager", **client_config)
        return self._client

    def _get_available_secrets(self) -> List[str]:
        all_secrets = self.client.list_secrets()

        return [
            secret["Name"]
            for secret in all_secrets["SecretList"]
            if f"{self.current_environment}/snatch" in secret["Name"]
        ]

    def get_project_secrets(self) -> Dict[Any, Any]:
        secrets_manager_data = {}
        needed_secrets = self._get_available_secrets()
        for secret in needed_secrets:
            logger.debug(f"Reading Secret: {secret} ...")
            secret_name = secret.split("/")[-1]
            secret_data = {secret_name: None}
            response = self.client.get_secret_value(SecretId=secret)
            data = json.loads(response["SecretString"])
            secret_data[secret_name] = {
                key.lower().strip(): data[key] for key in data.keys()
            }
            merge_dicts(secret_data, secrets_manager_data)
        return secrets_manager_data



import json
from pathlib import Path
from typing import Any, Dict


def load_fixture(filename: str) -> Dict[Any, Any]:
    """Load fixtures from test/fixtures folder.

    Example:
    * Add the file `my_response.json` in tests/fixtures
    * In tests load the file using::

        >>> from snatch.helpers.load_fixtures import load_fixture

        >>> def test_load_fixure():
        >>>     data = load_fixture("my_response.json")
        >>>     assert data is not None

    :param filename: JSON Filename
    :return: Dict
    """
    file_path = Path.cwd().joinpath("tests", "fixtures", filename)
    with open(str(file_path)) as file:
        fixture = json.load(file)
    return fixture



from datetime import timedelta
from functools import wraps
from typing import Callable

import arrow


def time_it(f: Callable) -> Callable:
    @wraps(f)
    def wrapper(*args, **kwargs) -> Callable:
        start_time = arrow.utcnow()
        data = f(*args, **kwargs)
        end_time = arrow.utcnow()
        delta = (end_time - start_time).microseconds
        data.timeit = str(timedelta(microseconds=delta))
        return data

    return wrapper



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['snatch', 'snatch.base', 'snatch.helpers']

package_data = \
{'': ['*']}

install_requires = \
['arrow',
 'asbool',
 'boto3',
 'loguru',
 'marshmallow',
 'python-dotenv',
 'requests',
 'stela',
 'validate-docbr']

setup_kwargs = {
    'name': 'py-snatch',
    'version': '1.2.0',
    'description': "The Friendly Integration Library for Data Scientists. Don't You Just Know It?",
    'long_description': None,
    'author': 'A55 Tech',
    'author_email': 'tech@a55.tech',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Optional, Type

from marshmallow import EXCLUDE, Schema, fields, post_load

from snatch.base.data import BaseData
from snatch.base.data_source import BaseSnatch
from snatch.base.schema import BaseSchema
from snatch.config import settings


@dataclass
class BoaVistaData(BaseData):
    """Boa Vista Data.

    Contains all data prospect from Boa Vista Data Source.

    Default Properties
    ==================
    * tax_id (str):                     The TaxId prospected.
                                        must be a valid CNPJ, formatted or not.
    * integration_status (str):         The current Integration Status, please check the
                                        `find_current_status` method for more info.
    * status_reason (str):              The human friendly reason for current status
    * prospect_date (datetime):         The prospect datetime.
    * payload (dict):                   Original payload received from prospection.
    * timeit                            Time in seconds for retrieving data.
    * datasource_base_url               Data Source Base Url

    Boa Vista Properties
    ====================
    * default_probability (float):      Default probability for TaxId.
    * score (int):                      Boa Vista Score for TaxId.
    * total_debts (float):              Boa Vista Total Debts registered.
    * tax_id_status (str):              TaxId status at Central Bank.
    """

    default_probability: Optional[float] = None
    score: Optional[int] = None
    total_debts: Optional[float] = None
    tax_id_status: Optional[str] = None


class BoaVistaSerializer(BaseSchema):
    default_probability = fields.Float(allow_none=True)
    score = fields.Integer(allow_none=True)
    total_debts: fields.Float(allow_none=True)
    tax_id_status = fields.Str(allow_none=True)
    payload = fields.Dict(allow_none=True)

    @post_load
    def make_data(self, data, **kwargs):
        new_data = BoaVistaData(**data)
        new_data.datasource_base_url = BoaVista.base_url
        return new_data

    class Meta:
        unknown = EXCLUDE


@dataclass
class BoaVista(BaseSnatch):
    default_timeout: int = 60
    default_max_days: int = 180
    base_url: str = settings["snatch.datasource_boa_vista_url"]
    authorization_token: str = settings["snatch.boa_vista_secret_token"]
    serializer_class: Type[Schema] = BoaVistaSerializer
